import numpy as np


def nms(boxes, scores, max_iou=0.3):
    '''
    Vectorized implementation of Non-Max Suppression.

    === Input ===
    boxes: array (N, 4)
        bounding boxes

    scores: array (N,)
        prediction scores / confidences

    max_iou: float
        maximum IoU allowed


    === Return ===
    keep_id: array of int
        index of bbox to keep
    '''
    pt1 = boxes[:, :2]
    pt2 = boxes[:, 2:]

    area = (pt2 - pt1).prod(1)
    idx = scores.argsort()

    keep_id = []
    while len(idx) > 0:
        # Keep most confident prediction
        best, other = idx[-1], idx[:-1]
        keep_id.append(best)

        # Calculate Intersection Area with best
        intersect_pt1 = np.maximum(pt1[best], pt1[other])
        intersect_pt2 = np.minimum(pt2[best], pt2[other])
        intersection = np.maximum(1e-8, intersect_pt2 - intersect_pt1).prod(1)

        # Calculate Intersection over Union (IoU)
        IoU = intersection / (area[best] + area[other] - intersection)

        # Filter bbox based on IoU
        idx = idx[:-1]
        idx = idx[IoU <= max_iou]
    return keep_id



import cv2


def key_pressed(key):
    """
    Listens to keyboard event


    === Input ===
    key: char
        keyboard key to listen to


    === Return ===
    A boolean value indicating the key being pressed
    """
    return cv2.waitKey(1) == ord(key)


def create_sized_window(height, aspect_ratio, window_name="window", window_type=cv2.WINDOW_NORMAL):
    '''
    Create a window with specified height and aspect ratio


    === Input ===
    height: int
        desired window height

    aspect_ratio: float
        media aspect ratio

    window_name: str
        window name

    window_type: enum
        Qt backend flags. These are the available options
        - cv2.WINDOW_NORMAL
            enables window resizing
        - cv2.WINDOW_AUTOSIZE
            automatically adjusts to fit the displayed image
        - cv2.WINDOW_OPENGL
            window with opengl support
        - cv2.WINDOW_FULLSCREEN
            change the window to fullscreen
        - cv2.WINDOW_FREERATIO
            adjusts the image with no respect to its ratio
        - cv2.WINDOW_KEEPRATIO
            keeps the image ratio
        - cv2.WINDOW_GUI_EXPANDED
            new enhanced GUI
        - cv2.WINDOW_GUI_NORMAL
            old way to draw the window without statusbar and toolbar


    === Return ===
    Create a new window, but returning None
    '''
    cv2.namedWindow(window_name, window_type)
    cv2.resizeWindow(window_name, int(height * aspect_ratio), height)



import cv2
import numpy as np
import matplotlib.pyplot as plt


def draw_bbox(frame, pt1, pt2, color, label=None, conf=None, thickness=1, font=cv2.FONT_HERSHEY_SIMPLEX, fontsize=0.5,
              fontcolor=(0, 0, 0), is_normalized=True):
    '''
    Draw a bounding box with its label


    === Input ===
    frame: array
        image / frame to be drawn

    pt1: (int, int) or (float, float)
        top left point of the bounding box.
        If is_normalized is True, then the point will be rescaled back based on the frame size.

    pt2: (int, int) or (float, float)
        bottom right point of the bounding box
        If is_normalized is True, then the point will be rescaled back based on the frame size.

    color: (int, int, int)
        The bounding box color in BGR format

    label: str
        A text or class label. It will be added to the inner top left of the bounding box

    conf: float
        The prediction confidence

    thickness: int
        The bounding box and text thickness

    font: opencv's font
        The text font. Check for the available font in opencv

    fontsize: float
        The font scaling factor towards the font's base size

    fontcolor: (int, int, int)
        The text font color in BGR format


    === Return ===
    frame: arrray
        annotated image / frame
    '''
    frame = frame.copy()
    if frame.dtype == np.float32:
        frame = (frame * 255).astype(np.uint8)

    if isinstance(color, int):
        color = tuple(c*255 for c in plt.cm.Set1(color)[:3])

    if is_normalized:
        h, w, c = frame.shape
        pt1 = pt1 * np.array([w, h])
        pt2 = pt2 * np.array([w, h])

    cv2.rectangle(frame, pt1, pt2, color, thickness)

    if label is not None:
        if conf is not None:
            text = f"{label} [{conf*100:.1f}%]"
        else:
            text = label

        # Handle text size
        (w_text, h_text), baseline = cv2.getTextSize(text, font, fontsize, thickness)

        # Filled textbox
        pt1_box = pt1
        pt2_box = (pt1[0] + w_text + 2 * baseline, pt1[1] + h_text + 2 * baseline)
        cv2.rectangle(frame, pt1_box, pt2_box, color, cv2.FILLED)

        # Add text
        pt_text = (pt1[0] + baseline, pt1[1] + h_text + baseline - thickness)
        cv2.putText(frame, text, pt_text, font, fontsize, fontcolor, thickness)
    return frame



import cv2
from jcopvision.exception import IncorrectExtensionError


class BaseWriter:
    def __init__(self, output_path, width, height, fps, fourcc):
        self._writer = cv2.VideoWriter(output_path, fourcc, fps, (int(width), int(height)))

    def write(self, frame):
        self._writer.write(frame)

    def close(self):
        self._writer.release()


class MP4Writer(BaseWriter):
    """
    A video writer with mp4 compression.
    """
    def __init__(self, output_path, width, height, fps):
        if not output_path.endswith(".mp4"):
            raise IncorrectExtensionError("output_path should have .mp4 extension")

        fourcc = cv2.VideoWriter_fourcc(*"MP4V")
        super().__init__(output_path, width, height, fps, fourcc)


class AVIWriter(BaseWriter):
    """
    A video writer with avi compression.
    """
    def __init__(self, output_path, width, height, fps):
        if not output_path.endswith(".avi"):
            raise IncorrectExtensionError("output_path should have .avi extension")

        fourcc = cv2.VideoWriter_fourcc(*"MJPG")
        super().__init__(output_path, width, height, fps, fourcc)




import cv2
import numpy as np
from os.path import splitext

from jcopvision.exception import MediaToArrayError


class MediaReader:
    """
    An all around media reader built on top of opencv.


    === Example Usage ===
    media = MediaReader("video.mp4")
    for frame in media.read():
        # do something
    media.close()
    """
    def __init__(self, source="webcam"):
        self.input_type, self.cam = self._parse_source(source)
        self._parse_prop()

    def _parse_source(self, source):
        _ext = lambda source: splitext(source)[-1]
        source = 0 if source == "webcam" else source

        if source == 0:
            input_type = "camera"
            cam = cv2.VideoCapture(source)
        elif _ext(source) in [".mp4", ".avi"]:
            input_type = "video"
            cam = cv2.VideoCapture(source)
        elif _ext(source) in [".bmp", ".dib", ".jpg", ".jpeg", ".jp2", ".jpe", ".png", ".pbm", ".pgm",
                              ".ppm", ".sr", ".ras", ".tiff", ".tif"]:
            input_type = "image"
            cam = cv2.imread(source)
        else:
            raise Exception("File type not supported")
        return input_type, cam

    def _parse_prop(self):
        if self.input_type == "image":
            h, w, c = self.cam.shape
            self.aspect_ratio = w / h
            self.height = int(h)
            self.width = int(w)
        else:
            self.height = int(self.cam.get(cv2.CAP_PROP_FRAME_HEIGHT))
            self.width = int(self.cam.get(cv2.CAP_PROP_FRAME_WIDTH))
            self.aspect_ratio = self.width / self.height
            self.frame_count = self.cam.get(cv2.CAP_PROP_FRAME_COUNT)
            self.frame_rate = self.cam.get(cv2.CAP_PROP_FPS)

    def read(self):
        if self.input_type == "image":
            return self.cam
        else:
            return self._video_reader()

    def close(self):
        if self.input_type in ['video', "camera"]:
            self.cam.release()

    def _video_reader(self):
        while True:
            cam_on, frame = self.cam.read()
            if cam_on:
                yield frame
            else:
                break

    def to_array(self, color_model="rgb"):
        if self.input_type == "video":
            frames = [frame for frame in self.read()]
            frames = np.array(frames).transpose(0, 3, 1, 2)
            if color_model == "rgb":
                frames = frames[:, ::-1, :, :]
            return frames
        else:
            raise MediaToArrayError("Image / webcam stream could not be converted to array. Input should be a video.")

